from .trainer_for_text_classification import TextClassificationTrainer
from .traniner_for_question_answer import QuestionAnswerTrainer
from .trainer_for_alpaca import AlpacaTrainer
from .trainer_for_llama_cn import Vicunatrainer
from .trainer_for_chinese_alpaca import ChineseAlpacaTrainer
from .trainer_for_baichuan import BaichuanTrainer
from .trainer_for_chatglm import ChatglmTrainer
from .trainer_for_ziya_llama_13b import Ziyallama13bTrainer
